"""
.. currentmodule:: neet.boolean

Boolean Networks
================

The :mod:`neet.boolean` module provides network types (:class:`WTNetwork` and
:class:`LogicNetwork`) and functions for simulating Boolean network modules.

API Documentation
-----------------
"""
from .wtnetwork import WTNetwork
from .logicnetwork import LogicNetwork
